#pragma once

#ifdef _WIN32
// From winerror.h
#undef NO_ERROR
#undef FAR
#undef NEAR
#endif

#include <array>
#include <vector>
#include <string>

namespace Trueface
{
    /** Error codes returned by methods. */
    enum class ErrorCode {NO_ERROR, INVALID_LICENSE, FILE_READ_FAIL, UNSUPPORTED_IMAGE_FORMAT, UNSUPPORTED_MODEL, NO_FACE_IN_FRAME, FAILED,
        COLLECTION_CREATION_ERROR, DATABASE_CONNECTION_ERROR, ENROLLMENT_ERROR, MAX_COLLECTION_SIZE_EXCEEDED, NO_RECORD_FOUND,
        NO_COLLECTION_FOUND, COLLECTION_DELETION_ERROR, EXTREME_FACE_ANGLE, FACE_TOO_CLOSE, FACE_TOO_FAR, FACE_TOO_SMALL,
        FACE_NOT_CENTERED, EYES_CLOSED, MASK_DETECTED, TOO_DARK, TOO_BRIGHT };

    /** Allow user to print ErrorCode for convenience. */
    inline std::ostream& operator<<(std::ostream & os, ErrorCode e) {
        switch (e) {
            case ErrorCode::NO_ERROR: return os << std::string("No error.");
            case ErrorCode::INVALID_LICENSE: return os << std::string("Error: License is not valid.");
            case ErrorCode::FILE_READ_FAIL: return os << std::string("Error: Unable to read file (ensure the filepath is set correctly).");
            case ErrorCode::UNSUPPORTED_IMAGE_FORMAT: return os << std::string("Error: Image format is not supported.");
            case ErrorCode::UNSUPPORTED_MODEL: return os << std::string("Error: Model is not supported yet.");
            case ErrorCode::NO_FACE_IN_FRAME: return os << std::string("Error: Unable to detect face in frame.");
            case ErrorCode::FAILED: return os << std::string("Error: There operation failed.");
            case ErrorCode::COLLECTION_CREATION_ERROR: return os << std::string("Error: Unable to create collection.");
            case ErrorCode::DATABASE_CONNECTION_ERROR: return os << std::string("Error: Unable to connect to database.");
            case ErrorCode::ENROLLMENT_ERROR: return os << std::string("Error: There was an error with enrollment.");
            case ErrorCode::MAX_COLLECTION_SIZE_EXCEEDED: return os << std::string("Error: The max collection size has been exceeded. Upgrade your token to support larger collections.");
            case ErrorCode::NO_RECORD_FOUND: return os << std::string("Error: Unable to find Faceprint for specified UUID for identity.");
            case ErrorCode::NO_COLLECTION_FOUND: return os << std::string("Error: Unable to find collection for specified name.");
            case ErrorCode::COLLECTION_DELETION_ERROR: return os << std::string("Error: Unable to delete collection.");
            case ErrorCode::EXTREME_FACE_ANGLE: return os << std::string("Error: Face angle is too extreme, use frontal facing image.");
            case ErrorCode::FACE_TOO_CLOSE: return os << std::string("Error: Face is too close.");
            case ErrorCode::FACE_TOO_FAR: return os << std::string("Error: Face is too far.");
            case ErrorCode::FACE_TOO_SMALL: return os << std::string("Error: Face is too small.");
            case ErrorCode::FACE_NOT_CENTERED: return os << std::string("Error: Face is not centered in the frame.");
            case ErrorCode::EYES_CLOSED: return os << std::string("Error: One or more eye's are closed. The eyes must be open.");
            case ErrorCode::MASK_DETECTED: return os << std::string("Error: A mask was detected. The face image must not be masked.");
            case ErrorCode::TOO_DARK: return os << std::string("Error: The image is too dark.");
            case ErrorCode::TOO_BRIGHT: return os << std::string("Error: The image is too bright.");
            default: return os << std::string("Unknown ErrorCode");
        }
    }

    /** Image color formats. */
    enum class ColorCode {bgr, rgb, bgra, rgba, gray, yuv_i420, yuv_nv12};

    /** Object classes. */
    enum class ObjectLabel {person, bicycle, car, motorcycle, airplane, bus, train, truck, boat, traffic_light,
        fire_hydrant, stop_sign, parking_meter, bench, bird, cat, dog, horse, sheep, cow,
        elephant, bear, zebra, giraffe, backpack, umbrella, handbag, tie, suitcase, frisbee,
        skis, snowboard, sports_ball, kite, baseball_bat, baseball_glove, skateboard, surfboard,
        tennis_racket, bottle, wine_glass, cup, fork, knife, spoon, bowl, banana, apple,
        sandwich, orange, broccoli, carrot, hot_dog, pizza, donut, cake, chair, couch,
        potted_plant, bed, dining_table, toilet, tv, laptop, mouse, remote, keyboard, cell_phone,
        microwave, oven, toaster, sink, refrigerator, book, clock, vase, scissors, teddy_bear,
        hair_drier, toothbrush};

    /** Filters the detected faces based on score thresholds obtained from ROC curve. */
    enum class FaceDetectionFilter {
        /**
         * Filter the detected faces based on a low score threshold.
         * Limits false negatives (does not detect a face),
         * but may have more false positives (classifies a non-face as a face). */
        HIGH_RECALL,
        /**
         * Filter the detected faces based on a high score threshold.
         * Limits false positives (classifies a non-face as a face),
         * but may have more false negatives (does not detect a face). */
        HIGH_PRECISION,
        /**
         * Filter the detected faces based on a medium score threshold
         * to balance false positives and false negatives.
         * We advise using this option most of the time.
         * */
        BALANCED,
        /**
         * Do not filter the detected faces by score.
         * Will have a large number of false positives (classifies a non-face as a face). */
        UNFILTERED
    };

    /** Database Management System for storing Faceprints. **/
    enum class DatabaseManagementSystem {
        /** Use sqlite backend.
         * Write Faceprints to local disk.
         * Ideal for embedded systems or use cases where only one process connects to the database.*/
        SQLITE,
        /** Use a PostgreSQL backend.
         * Ideal for distributed systems requiring synchronization. */
        POSTGRESQL,
        /** Do not write Faceprints to disk, only store in ram.
         * Warning, enrolled Faceprints will not be saved after the program terminates.
         * Switching to a new collections will also delete all enrolled templates. */
        NONE
    };

    /** Facial recognition models.
     * To compare model performances, refer to our <a href="https://docs.trueface.ai/roc-curves">ROC curves</a>.
     * You can also find more information on our <a href="https://reference.trueface.ai/cpp/dev/latest/py/faq.html#what-are-the-differences-between-the-face-recognition-models">FAQ page</a>.
     * You can view the inference time for the various models on our <a href="https://docs.trueface.ai/benchmarks">benchmarks page</a>.
     * The current most accurate model is TFV5. */
    enum class FacialRecognitionModel {
        /** Our most lightweight model with fastest inference time but lowest accuracy,
         * ideal for embedded systems or lightweight CPU only deployments, prototyping, and some 1 to 1 matching use cases.*/
        LITE,
        /**
         * Lightweight model which has improved accuracy over the previous LITE model, though does have slightly greater inference time.
         * Ideal for embedded systems or lightweight CPU only deployments, prototyping, and some 1 to 1 matching use cases.
         */
        LITE_V2,
        /** Full TFV4 model which has better accuracy than the LITE model, but also has greater inference time.
         * Ideal for GPU deployments and for 1 to N use cases.
         * Note, TFV4 has now been deprecated and replaced by TFV5 which has better performance.
         * Despite this, we will continue providing support for TFV4 for clients with existing collections. */
        FULL [[deprecated("Use TFV5 instead.")]],
        /**
         * TFV5 is currently the highest accuracy model for unmasked face images.
         * Ideal for GPU deployments and for 1 to N use cases. */
        TFV5,
        /**
         * TFV6 is currently the highest accuracy model for masked face images.
         * Use TFV6 in situations where it is anticipated that the probe image contains a masked face (for 1 to N search),
         * or where one or both face images are masked (for 1 to 1 comparisons).
         * TFV6 has comparable inference time to TFV5.
         */
        TFV6
    };

    /**
     * Object detection models.
     * */
    enum class ObjectDetectionModel {
        /**
         * Resizes the input image to 1280x1280 (uses letterbox padding to maintain aspect ratio).
         * Should be used for image where one or both dimensions are greater than 1280, and images with small objects.
         */
        ACCURATE,
        /**
         * Resizes the input image to 640x640 (uses letterbox padding to maintain aspect ratio).
         * Should be used for smaller images, or images with large objects.
         */
        FAST
    };

    /** Precision to use for GPU inference. */
    enum class Precision {
        /**
         * 32 bit floating point.
         * Allows for highest accuracy but slower inference.
         */
        FP32,
        /**
         * 16 bit floating point.
         * Allows for faster inference but lower accuracy.
         * Only supported by TFV6 and newer models.
         */
        FP16
    };

    /** GPU options for a specific module (ex. face detector). */
    struct GPUModuleOptions {
        /** Precision level used for inference. */
        Precision precision = Precision::FP16;
        /** The maximum batch size which will be used. */
        int32_t maxBatchSize = 4;
        /**
         * The batch size which should be optimized for.
         * Must be less than or equal to `Trueface::GPUModuleOptions.maxBatchSize`.
         */
        int32_t optBatchSize = 1;
        /**
         * The maximum allowable GPU memory to be used for model conversion, in Mb.
         * Applications should allow the engine builder as much workspace as they can afford.
         * At runtime, the SDK allocated no more than this and typically less.
         */
        size_t maxWorkspaceSizeMb = 2000;
    };

    /**
     * GPU options for the SDK.
     * Note, GPU support requires a different version of the SDK.
     * Default uses CPU for inference. */
    struct GPUOptions {
        /** Enable GPU inference for all supported modules. **/
        bool enableGPU = false;
        /** GPU device index. */
        unsigned int deviceIndex = 0;
        /** Options for face detector GPU inference. */
        GPUModuleOptions faceDetectorGPUOptions {};
        /** Options for face recognizer GPU inference. */
        GPUModuleOptions faceRecognizerGPUOptions {};
        /** Options for mask detector GPU inference. */
        GPUModuleOptions maskDetectorGPUOptions {};
        /**
         * Options for object detector GPU inference.
         * Note, object detector only supports a batch size of 1 at this moment.
         * */
        GPUModuleOptions objectDetectorGPUOptions {};

        GPUOptions() = default;
        /** Enable or disable GPU inference for all supported modules. */
        GPUOptions(bool val) : enableGPU(val) {};
    };

    /**
     * Encrypt the biometric templates and identity strings when storing in the database using AES encryption.
     * Note, enabling this option does add overhead to Faceprint enrollment as well as loading a collection from a database into memory.
     * Enabling encryption does not increase the 1 to N identification time.
     */
    struct EncryptDatabase {
        /**
         * Enable database encryption.
         * Must provide encryption key if encryption is enabled.
         * If enabling encryption with PostgreSQL backend, it is strongly advised to require SSL for PostgreSQL connection.
         **/
        bool enableEncryption = false;
        /**
         * Encryption key.
         * The key is hashed to a fixed length before being used for encryption.
         * */
        std::string key;
    };

    /**
     * Initialize module in SDK constructor.
     * By default, the SDK uses lazy initialization, meaning modules are only initialized when they are first used (on first inference).
     * This is done so that modules which are not used do not load their models into memory, and hence do not utilize memory.
     * The downside to this is that the first inference will be much slower as the model file is being decrypted and loaded into memory.
     * Therefore, if you know you will use a module, choose to pre-initialize the module, which reads the model file into memory in the SDK constructor.
     */
    struct InitializeModule {
        /** Face detector. */
        bool faceDetector = false;
        /** Face recognizer. */
        bool faceRecognizer = false;
        /** Object detector. */
        bool objectDetector = false;
        /** Bodypose estimator. */
        bool bodyposeEstimator = false;
        /** Liveness. */
        bool liveness = false;
        /** Active spoof. */
        bool activeSpoof = false;
        /** Passive spoof. */
        bool passiveSpoof = false;
        /** 106 face point landmark detector. */
        bool landmarkDetector = false;
        /** Mask Detector. */
        bool maskDetector = false;
    };

    struct UninitializeModule {
        bool maskDetector = false;
        bool faceRecognizer = false;

        bool faceDetector = false;
        bool objectDetector = false;
        bool bodyposeEstimator = false;
        bool liveness = false;
        bool activeSpoof = false;
        bool passiveSpoof = false;
        bool landmarkDetector = false;
    };

    /** SDK configuration options. */
    struct ConfigurationOptions {
        /**
         * **Android Only**
         * Power saving mode.
         * 0 = all cores enabled (default).
         * 1 = only little clusters enabled.
         * 2 = only big clusters enabled.
        */
        int mobilePowerSave = 0;
        /**
         * **Mobile Only**
         * Set the number of threads used for inference.
         * For non-mobile platforms, should use the ``OMP_NUM_THREADS`` environment variable.
         * */
        int mobileThreads = 4;
        /**
         * **iOS Only**
         * SDK user must define how much memory is available for iOS application.
         * Should be set using the following code:
         * #include <os/proc.h>
         * options.mobileAvailableMemory = os_proc_available_memory();
         * */
        size_t mobileAvailableMemory = 0;

        /** The model to be used for facial recognition (default is TFV5). */
#if __APPLE__
        #include "TargetConditionals.h"
    #if TARGET_OS_IPHONE && TARGET_OS_SIMULATOR
        FacialRecognitionModel frModel = FacialRecognitionModel::LITE;
    #else
        FacialRecognitionModel frModel = FacialRecognitionModel::TFV5;
    #endif
#else
        FacialRecognitionModel frModel = FacialRecognitionModel::TFV5;
#endif

        /** The model to be used for object detection (default is ACCURATE model). */
        ObjectDetectionModel objModel = ObjectDetectionModel::ACCURATE;
        /**
         * The smallest face height that the face detector can detect. (default is 40 pixels, min value is 16 pixels).
         * The face detector has a detection scale range of about 5 octaves.
         * Ex. 40 pixels yields the detection scale range of ~40 pixels to 1280 (=40x2^5) pixels.
         * If set to -1, will dynamically adjusts the face detection scale range from image-height/32 to image-height
         * to ensure that large faces are detected in high resolution images.
         * Increasing the `Trueface::ConfigurationOptions.smallestFaceHeight` will result in faster face detection.
        **/
        int smallestFaceHeight = 40;
        /**
        * The face detection filter (default is BALANCED).
        * Filters the detected faces based on score thresholds obtained from ROC curve. */
        FaceDetectionFilter fdFilter = FaceDetectionFilter::BALANCED;
        /** Database management system for storing Faceprints (default is SQLITE). **/
        DatabaseManagementSystem dbms = DatabaseManagementSystem::SQLITE;
        /** The directory path containing the model files. **/
        std::string modelsPath = "./";
        /** Improves the computation speed for 1 to 1 comparisons and 1 to N searches
         * by compressing the feature vector and enabling additional optimizations. **/
        bool frVectorCompression = false;

        /**
         * Options for enabling and configuring GPU inference.
         * Default uses CPU inference.
         * Note, GPU support requires a different version of the SDK.
         * You can easily enable GPU inference for all supported modules by setting this option to true.
        **/
         GPUOptions gpuOptions = GPUOptions{false};

        /**
         * Though this option is deprecated, it should continue to be used for older models such as the FULL and TFV5 models.
         * For newer models such as TFV6 and newer, use Trueface::GPUModuleOptions.optBatchSize instead.
         * Specify the batch sizes which will be used, resulting in improved performance when switching between batch sizes.
         * If the batch sizes are specified, then using a non-specified batch size will result in an exception being thrown.
         * Note, GPU memory will be allocated for each of the specified batch sizes,
         * so specifying too many batch sizes may result in an out-of-memory crash.
         * Leave vector empty to support dynamic batch sizes. Dynamic batch sizes will result in a slowdown when switching between batch sizes.
         * Ex. If processing 36 images in the following batches: 10, 10, 10, 6 there will be a slight slowdown when switching from the batch size of 10 to 6.
         * */
        [[deprecated("Use Trueface::GPUModuleOptions.optBatchSize instead.")]]
        std::vector<unsigned int> batchSizes = {};
        /** Initialize specified modules in the SDK constructor (default uses lazy initialization). */
        InitializeModule initializeModule{};
        /** Encrypt the biometric templates and identity strings when storing in the database using AES encryption (default is disabled). */
        EncryptDatabase encryptDatabase{};
    };

    template <typename T>
    struct Point {
        /** Coordinate along the horizontal axis, or pixel column. */
        T x;
        /** Coordinate along the vertical axis, or pixel row. */
        T y;
    };

    /** Options used when generating the feature vector. **/
    struct ModelOptions {
        /** Indicates if the `Trueface::ConfigurationOptions.frVectorCompression` was enabled when generating the template. **/
        bool frVectorCompression = false;
    };

    /** Face feature vector and meta data.  **/
    struct Faceprint {
        /** Vector of floats which describe the face. **/
        std::vector<float> featureVector;
        /** SDK version used to generate feature vector. **/
        std::string sdkVersion;
        /** Name of model used to generate feature vector. **/
        std::string modelName;
        /** Additional options used when generating the feature vector. **/
        ModelOptions modelOptions;
    };

    struct BoundingBox {
        /** The object class as a label. */
        ObjectLabel label;
        /** The detection's confidence probability. */
        float probability;
        /** The top left corner Point of the bounding box. */
        Point<float> topLeft;
        /**  The width of the bounding box. */
        float width;
        /**  The height of the bounding box. */
        float height;
    };

    /** Landmark representing body joint. */
    struct Landmark {
        /** Coordinate of the joint. */
        Point<float> point;
        /** Score for the joint. */
        float score;
        /** Label for the joint. */
        std::string label;
    };

    struct FaceBoxAndLandmarks
    {
        /**  The top left corner Point of the bounding box. */
        Point<float> topLeft;
        /**  The bottom right corner Point of the bounding box. */
        Point<float> bottomRight;
        /**  A vector of Point representing the facial landmarks: subject right eye, subject left eye, nose, subject right mouth corner, subject left mouth corner. */
        std::vector<Point<float>> landmarks;
        /** Likelihood of this being a true positive; a value lower than 0.85 indicates a high chance of being a false positive. */
        float score;
    };

    /** Array of 106 face points. */
    using Landmarks = std::array<Point<float>, 106>;

    struct Candidate {
        /** The computed similarity measure. */
        float similarityMeasure;
        /** The probability the two face feature vectors are a match. */
        float matchProbability;
        /** The identity of the match. */
        std::string identity;
        /** The UUID of the matching Faceprint. */
        std::string UUID;
    };

    /** Results from spoof detection. */
    enum class SpoofLabel {
        /** The face image is real. */
        REAL,
        /** The image may be an attempted spoof. */
        FAKE
    };

    /**
     * Active spoof involves two stages: a near image and a far image.
     * The following enum is used to track which stage image is being analyzed.
     */
    enum class ActiveSpoofStage {
        /** Near image. */
        /** Near image. */
        NEAR,
        /** Far image. */
        FAR
    };

    /** Results from mask detection. */
    enum class MaskLabel {
        /** There is a mask on the face. */
        MASK,
        /** There is no mask on the face. */
        NO_MASK
    };

    /** Results from eye glasses detection. */
    enum class GlassesLabel {
        /** There are glasses on the face. */
        GLASSES,
        /** There are no glasses on the face. */
        NO_GLASSES
    };

    /** Collection metadata. */
    struct CollectionMetadata {
        /** The name of the collection. */
        std::string collectionName = "";
        /** The number of unique identities in the collection. */
        uint64_t numIdentities = 0;
        /** The total number of Faceprint enrolled in the collection. */
        uint64_t numFaceprints = 0;
        /** The name of the face recognition model used to generate the Faceprint enrolled in the collection. */
        std::string modelName = "";
        /** The size of the Faceprint feature vector in bytes. */
        uint16_t featureVectorSizeBytes = 0;
        /** Additional options which were used to generate Faceprint enrolled in the collection. */
        ModelOptions modelOptions {};
        /** Indicates if the collection is encrypted. */
        bool encrypted = false;
    };

    /**
     * Results from blink detection.
     * You can use the members of `Trueface::BlinkState` to set your own thresholds which work best for your camera and lighting conditions.
     * */
    struct BlinkState {
        /** Predicted result for the left eye being closed.
         * Computed as `Trueface::BlinkState.leftEyeScore` < 0.3 && `Trueface::BlinkState.leftEyeAspectRatio` < 0.22. */
        bool isLeftEyeClosed = false;
        /** Predicted result for the right eye being closed.
         * Computed as `Trueface::BlinkState.rightEyeScore` < 0.3 && `Trueface::BlinkState.rightEyeAspectRatio` < 0.22. */
        bool isRightEyeClosed = false;
        /** Score indicating if the left eye is open or closed.
         * 0 indicates closed, 1 indicates open. */
        float leftEyeScore = 0.f;
        /** Score indicating if the right eye is open or closed.
         * 0 indicates closed, 1 indicates open. */
        float rightEyeScore = 0.f;
        /** The aspect ratio of the left eye, computed as height / width. */
        float leftEyeAspectRatio = 0.f;
        /** The aspect ratio of the right eye, computed as height / width. */
        float rightEyeAspectRatio = 0.f;

    };

    /** Flags used for rotating images. */
    enum class RotateFlags {
        /** Rotate 90 degrees clockwise. */
        ROTATE_90_CLOCKWISE,
        /** Rotate 180 degrees clockwise. */
        ROTATE_180,
        /** Rotate 270 degrees clockwise. */
        ROTATE_90_COUNTERCLOCKWISE
    };
}